"""
Tries to figure out the file type based on content.
"""

import csv
import re


def magic(file_path):
    """
    Based on the first few blocks of `file_path` contents, figure out the likely
    data file serialization type.
    """
    # if XML, it's likely rdf/xml
    # if JSON, it's likely json-ld
    # we go through the various text-based formats before finally seeing if we can parse it
    # as CSV or similar (regular separator character, line feeds, etc.)
    with open(file_path, "r") as file:
        blocks = file.read(
            8096
        )  # read 8k - should be enough to figure out what we have
        if re.match(r"^\s*<\?xml", blocks):
            return "application/rdf+xml"
        if re.match(r"^\s*<!--", blocks):
            return "application/rdf+xml"
        if re.match(r"^\s*<[a-zA-Z]+://", blocks):
            return "application/n-triples"
        if re.match(r"^\s*<\S", blocks):
            return "application/rdf+xml"
        if re.match(r"^\s*@(base|prefix)\s+", blocks):
            return "text/turtle"
        if re.match(r"^\s*(BASE|PREFIX)\s+", blocks):
            return "text/turtle"
        if re.match(r"^\s*[{[]", blocks):
            return "application/ld+json"
        if _maybe_csv(blocks):
            return "text/csv"
        return "text/plain"


def _maybe_csv(blocks):
    """
    Try to figure out if the file is likely to be CSV or similar.
    """
    dialect = csv.Sniffer().sniff(blocks)
    return dialect is not None



import sys

from cement import Controller, ex

from ..core.services import client


class Show(Controller):
    """
    Provides the `ursactl show ...` set of commands.
    """

    class Meta:
        label = "show"
        stacked_on = "base"
        stacked_type = "nested"
        help = "show something"

    @ex(help="show platform endpoint")
    def platform(self):
        print(self.app.config.get("ursactl", "platform"))

    @ex(help="show default project")
    def project(self):
        print(self.app.config.get("ursactl", "project"))

    @ex(help="show token")
    def token(self):
        iam_client = client("iam", self.app)
        token = iam_client.get_token()
        if token:
            print(token)
        else:
            sys.exit(1)



import asyncio
from concurrent.futures import FIRST_COMPLETED


class DONE:
    pass


async def get_unless_done(coro, done_future):
    getter_future = asyncio.ensure_future(coro)
    done, pending = await asyncio.wait(
        [getter_future, done_future], return_when=FIRST_COMPLETED
    )
    if done_future in done:
        if not getter_future.done():
            (getter,) = pending
            getter.cancel()
        return DONE
    return await getter_future



import json

from ursactl.core._base import Base
from ursactl.core.services import client

from .running_agent import RunningAgent


class Agent(Base):
    """
    Provides access to an agent definition.
    """

    @property
    def client(self):
        if self._client is None:
            self._client = client("planning", self.app)
        return self._client

    @property
    def name(self):
        return self._data["name"]

    @property
    def packages(self):
        return self._data["packages"]

    @property
    def behaviors(self):
        return self._data["behaviors"]

    def run(self, state=None, predicates=None):
        result = self.client.run_agent(
            self.uuid,
            initial_state=json.dumps(state),
            initial_predicates=json.dumps(predicates),
            project_uuid=self.project.uuid,
        )

        if result["errors"]:
            return None
        running_agent_id = result["result"]["id"]
        return RunningAgent(
            uuid=running_agent_id,
            app=self.app,
            client=self.client,
            project=self.project,
        )

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    "name": None,
                    "packages": None,
                    "behaviors": None,
                    "project_uuid": None,
                }
            else:
                self._cached_data = self.client.get_agent(self.uuid)
        return self._cached_data



from ursactl.core._base import Base
from ursactl.core.services import client


class RunningAgent(Base):
    """
    Provides access to a running agent.
    """

    @property
    def client(self):
        if self._client is None:
            self._client = client("planning", self.app)
        return self._client

    def send_event(self, domain, name, params):
        result = self.client.send_event_to_agent(
            self.uuid, {"domain": domain, "name": name, "params": params}
        )
        return result["result"]

    def test_event(self, domain, name, params):
        result = self.client.test_event_with_agent(
            self.uuid, {"domain": domain, "name": name, "params": params}
        )
        return result

    def send_events(self, events):
        actions = []
        for event in events:
            result = self.client.send_event_to_agent(self.uuid, event)
            if result["result"]:
                actions.extend(result["result"])
        return actions

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.stop_agent(self.uuid)



def get_provider_info():
    return {
        "package-name": "ursactl",
        "name": "Ursa Frontier",
        "description": "Provides access to Ursa Frontier SaaS Platform datasets and data generation pipelines.",
        "connection-types": [
            {
                "connection-type": "ursa_frontier_platform",
                "hook-class-name": "ursactl.airflow_provider.hooks.UrsaCtlHook",
            }
        ],
    }



from cement import Controller
from cement.utils.version import get_version_banner

from ..core.version import get_version

VERSION_BANNER = """
Ursa Frontier CLI. %s
%s
""" % (
    get_version(),
    get_version_banner(),
)


class Base(Controller):
    class Meta:
        label = "base"

        title = "verbs"

        # text displayed at the top of --help output
        description = (
            "Command line control script and library for Ursa Frontier SaaS Platform."
        )

        # text displayed at the bottom of --help output
        epilog = "Usage: ursactl verb ..."

        # controller level arguments. ex: 'ursactl --version'
        arguments = [
            # add a version banner
            (["-v", "--version"], {"action": "version", "version": VERSION_BANNER}),
            (
                ["-o", "--output"],
                {
                    "help": "Select an alternate output format (json, yaml)",
                    "action": "store",
                    "dest": "output",
                },
            ),
            (
                ["--connect-timeout"],
                {
                    "help": "Connect timeout in seconds (float)",
                    "dest": "connect_timeout",
                    "action": "store",
                },
            ),
            (
                ["--read-timeout"],
                {
                    "help": "Read timeout in seconds (float)",
                    "dest": "read_timeout",
                    "action": "store",
                },
            ),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()



"""
Manages service discovery.
"""

from typing import Optional

from .ape_client import ApeClient
from .async_client.client import AsyncClient
from .dss_client import DssClient
from .iam_client import IamClient
from .page_client import PageClient
from .planning_client import PlanningClient
from .project_client import ProjectClient

SERVICE_CLIENTS = {
    "ape": ApeClient,
    "dss": DssClient,
    "page": PageClient,
    "iam": IamClient,
    "planning": PlanningClient,
    "projects": ProjectClient,
    "async": AsyncClient,
}


def client(service: Optional[str] = None, app=None):
    """
    Returns a client appropriate for the service. The optional `app` parameter
    allows the reuse of an UrsaCtl object to manage configuration.
    """
    if app is None:
        from ursactl.main import UrsaCtl

        app = UrsaCtl()
        app.reload()

    platform = app.config.get("ursactl", "platform")

    if service in SERVICE_CLIENTS:
        return SERVICE_CLIENTS[service](platform, app)



"""
IAM API client.
"""

import configparser
import sys

import requests
from requests.auth import HTTPBasicAuth
from tinydb import Query


class IamClient:
    """
    A client for accessing IAM functions in the Ursa Frontier platform.

    This is limited to using API credentials to get a token that can be
    used with other platform services.
    """

    def __init__(self, endpoint: str, app):
        self.app = app
        self.endpoint = endpoint

    def clear_token(self):
        """
        Used to remove a cached token when it doesn't work.
        """
        try:
            api_key = self.app.config.get("ursactl", "api_key")
        except configparser.NoOptionError:
            print(
                "You need to set the api_key and api_secret in the ursactl.conf file."
            )
            sys.exit(1)

        query = Query()
        self.app.db.remove(query.api_key == api_key)

    def get_token(self):
        """
        Retrieves a token that can be used with other services.
        """
        try:
            api_key = self.app.config.get("ursactl", "api_key")
            api_secret = self.app.config.get("ursactl", "api_secret")
        except configparser.NoOptionError:
            print(
                "You need to set the api_key and api_secret in the ursactl.conf file."
            )
            sys.exit(1)

        # check if it's in the tinydb -- if so, return it
        query = Query()
        results = self.app.db.search(query.api_key == api_key)

        if results:
            return results[0]["token"]

        # otherwise, get a token, save it to tinydb, and then return it
        response = requests.get(
            f"{self.endpoint}api/token",
            auth=HTTPBasicAuth(api_key, api_secret),
            timeout=self.timeout,
        )

        if response.status_code != 200:
            return

        token = response.json()["token"]
        self.app.db.insert({"api_key": api_key, "token": token})
        return token

    @property
    def timeout(self):
        return (
            self._get_arg("connect_timeout", 60.0),
            self._get_arg("read_timeout", 60.0),
        )

    def _get_arg(self, key, default):
        try:
            pargs = vars(self.app.pargs)
            value = pargs[key]
        except AttributeError:
            value = None
        except TypeError:
            value = None

        if value is None:
            try:
                value = self.app.config.get("ursactl", key)
            except configparser.NoOptionError:
                value = default
        return float(value)



"""
Page Service client
"""

from ursactl.core.exc import UrsaBadProjectName

from ._base import Base

LIST_PAGES_QUERY = """
query list_pages($id: String!) {
    project(handleName: $id) {
        pages {
            id
            title
            path
            isSynced
            tags
        }
    }
}
"""

CREATE_PAGE_MUTATION = """
mutation create_page($input: CreatePageInput!) {
    createPage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            title
            path
        }
    }
}
"""

SYNC_PAGE_MUTATION = """
mutation sync_page($input: SyncPageInput!) {
    syncPage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            title
            path
        }
    }
}
"""

UPDATE_PAGE_MUTATION = """
mutation update_page($id: ID!, $input: UpdatePageInput!) {
    updatePage(id: $id, input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            title
            path
        }
    }
}
"""

GET_PAGE_BY_PATH_QUERY = """
query get_page_by_path($path: String!, $project: String!) {
    project(handleName: $project) {
        page(limit: 1, filter: { path: { eq: $path } }) {
            id
            path
            isSynced
            tags
        }
    }
}
"""

GET_PAGE_QUERY = """
query get_page($id: ID!) {
    page(id: $id) {
        id
        path
        title
        isSynced
        tags
    }
}
"""

GET_PAGE_CONTENT_QUERY = """
query get_page_content($id: ID!) {
    page(id: $id) {
        id
        path
        title
        content
        isSynced
        tags
    }
}
"""

DELETE_PAGE_MUTATION = """\
mutation delete_page($id: ID!) {
    deletePage(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_SYNCED_PAGE_MUTATION = """\
mutation delete_synced_page($id: ID!) {
    deleteSyncedPage(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""


class PageClient(Base):
    """
    Client providing access to the Page functions.
    """

    def list_pages(self, project_uuid=None):
        """
        Returns a list of available pages.
        """
        result = self.raw_query(LIST_PAGES_QUERY, {"id": project_uuid})
        if result["data"]["project"] is None:
            raise UrsaBadProjectName(project_uuid)
        return result["data"]["project"]["pages"]

    def get_page(self, uuid=None, path=None, project_uuid=None):
        """
        Returns more detail about a page - but not the content.
        """
        if uuid is None and path is None:
            raise ValueError("Must specify either uuid or path")
        if uuid is None:
            result = self.raw_query(
                GET_PAGE_BY_PATH_QUERY, {"project": project_uuid, "path": path}
            )

            return result["data"]["project"]["page"]

        result = self.raw_query(GET_PAGE_QUERY, {"id": uuid})
        return result["data"]["page"]

    def create_page(
        self, project_uuid=None, title=None, path=None, content=None, tags=None
    ):
        """
        Create a new page.
        """
        if project_uuid is None:
            raise AttributeError("pages must be assigned to a project.")
        if path is None:
            raise AttributeError("page path must be provided.")
        if content is None:
            raise AttributeError("page content must be provided.")
        variables = {
            "input": {
                "project": project_uuid,
                "path": path,
                "content": content,
            }
        }

        if title is not None:
            variables["input"]["title"] = title
        if tags is not None:
            variables["input"]["tags"] = tags
        result = self.raw_query(query=CREATE_PAGE_MUTATION, variables=variables)

        return result["data"]["createPage"]["result"]

    def sync_page(
        self, project_uuid=None, title=None, path=None, content=None, tags=None
    ):
        """
        Sync a page.
        """
        if project_uuid is None:
            raise AttributeError("pages must be assigned to a project.")
        if path is None:
            raise AttributeError("page path must be provided.")
        if content is None:
            raise AttributeError("page content must be provided.")
        variables = {
            "input": {
                "project": project_uuid,
                "path": path,
                "content": content,
            }
        }

        if title is not None:
            variables["input"]["title"] = title
        if tags is not None:
            variables["input"]["tags"] = tags
        result = self.raw_query(query=SYNC_PAGE_MUTATION, variables=variables)

        return result["data"]["syncPage"]["result"]

    def update_page(self, uuid=None, title=None, path=None, content=None, tags=None):
        """
        Update a page.
        """
        if uuid is None:
            raise AttributeError("page uuid must be provided.")
        variables = {"id": uuid, "input": {}}
        if title is not None:
            variables["input"]["title"] = title
        if path is not None:
            variables["input"]["path"] = path
        if content is not None:
            variables["input"]["content"] = content
        if tags is not None:
            variables["input"]["tags"] = tags
        result = self.raw_query(query=UPDATE_PAGE_MUTATION, variables=variables)

        return result["data"]["updatePage"]["result"]

    def delete_page(self, id, project_uuid=None):
        """
        Delete a page.
        """
        if self.is_uuid(id):
            variables = {"id": id}
        else:
            page = self.get_page(path=id, project_uuid=project_uuid)
            if page is None:
                return {}
            variables = {"id": page["id"]}

        result = self.raw_query(query=DELETE_PAGE_MUTATION, variables=variables)

        return result["data"]["deletePage"]

    def delete_synced_page(self, id, project_uuid=None):
        """
        Delete a synced page.
        """
        if self.is_uuid(id):
            variables = {"id": id}
        else:
            page = self.get_page(path=id, project_uuid=project_uuid)
            if page is None:
                return {}
            variables = {"id": page["id"]}

        result = self.raw_query(query=DELETE_SYNCED_PAGE_MUTATION, variables=variables)

        return result["data"]["deleteSyncedPage"]



"""
Base class for API clients.
"""

import configparser
import re

import requests
from python_graphql_client import GraphqlClient

from ursactl.core.exc import UrsaNotAuthorized


class Base:
    """
    Base class for clients.
    """

    def __init__(self, endpoint: str, app):
        from .. import services

        self.app = app
        self.endpoint = endpoint
        self.iam_client = services.client("iam", app)
        self.client = GraphqlClient(endpoint=f"{endpoint}api", timeout=self.timeout)

    def raw_query(self, query, variables=None, reauthorize=True):
        if variables is None:
            variables = {}

        try:
            result = self.client.execute(
                query=query,
                variables=variables,
                headers={"authorization": f"Bearer {self.iam_client.get_token()}"},
            )

            if "errors" in result:
                for error in result["errors"]:
                    if "code" in error and error["code"] == "Forbidden":
                        if reauthorize:
                            self.iam_client.clear_token()
                            return self.raw_query(query, variables, reauthorize=False)
                        raise UrsaNotAuthorized("Not authorized")

                    print(error["message"])
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                if reauthorize:
                    self.iam_client.clear_token()
                    return self.raw_query(query, variables, reauthorize=False)
                raise UrsaNotAuthorized("Not authorized") from e

            raise e
        return result

    @property
    def timeout(self):
        return (
            self._get_arg("connect_timeout", 60.0),
            self._get_arg("read_timeout", 60.0),
        )

    def _get_arg(self, key, default):
        try:
            pargs = vars(self.app.pargs)
            value = pargs[key]
        except AttributeError:
            value = None
        except TypeError:
            value = None

        if value is None:
            try:
                value = self.app.config.get("ursactl", key)
            except configparser.NoOptionError:
                value = default
        return float(value)

    @staticmethod
    def is_uuid(s: str) -> bool:
        """
        Test if the string looks like a UUID.
        """
        if s is None:
            return False
        return re.match(
            r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", s
        )



"""
DataStore Service client
"""

import hashlib
import urllib.parse

import requests

from ursactl.core.exc import UrsaBadProjectName, UrsaNotAuthorized

from ._base import Base

UPDATE_DATASET_MUTATION = """\
mutation update_dataset($id: ID!, $input: UpdateDatasetInput!) {
    updateDataset(id: $id, input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

GET_DATASET_PATH_QUERY = """\
query get_dataset($id: ID!) {
    dataset(id: $id) {
        path
        project { id name user { handle } }
    }
}
"""

GET_DATASET_QUERY = """\
query get_dataset($id: ID!) {
    dataset(id: $id) {
        id
        path
        contentType
        description
        size
        isSynced
        project { id }
    }
}
"""

GET_DATASET_BY_PATH_QUERY = """\
query get_dataset($projectId: String!, $path: String!) {
    project(handleName: $projectId) {
        datasets(filter: { path: { eq: $path } }, limit: 1) {
            id
            path
            contentType
            description
            size
            project { id }
        }
    }
}
"""

DELETE_DATASET_MUTATION = """\
mutation delete_dataset($id: ID!) {
    deleteDataset(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_SYNCED_DATASET_MUTATION = """\
mutation delete_synced_dataset($id: ID!) {
    deleteSyncedDataset(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

LIST_DATASETS_FOR_PROJECT_QUERY = """\
query listDatasets($projectId: String!) {
    project(handleName: $projectId) {
        datasets(limit:100) {
            id
            path
            size
            contentType
            isSynced
        }
    }
}
"""


class DssClient(Base):
    @staticmethod
    def _raise_if_none(value, var_name):
        if value is None:
            raise ValueError(f"{var_name} must be provided")

    def create_dataset(
        self,
        project_uuid=None,
        path=None,
        description=None,
        data=None,
        data_type=None,
        reauthorize=True,
    ):
        """
        Creates a dataset, returning the uuid and id of the dataset.

        If no data is provided, the dataset is created with no attached data.
        Data can be attached later.
        """
        # we first use the graphql to create the dataset
        # then we use a PUT to upload the data
        # the Python graphql client doesn't support multi-part HTTP requests
        self._raise_if_none(project_uuid, "project_uuid")
        self._raise_if_none(path, "path")
        self._raise_if_none(data, "data")
        self._raise_if_none(data_type, "data_type")

        # first, we POST to the right URL to upload the contents of the file
        with open(data, "rb") as file:
            contents = file.read()
        res = requests.post(
            url=f"{self.endpoint}api/dss/{project_uuid}/{path}",
            data=contents,
            headers={
                "Content-Type": data_type,
                "Authorization": f"Bearer {self.iam_client.get_token()}",
            },
        )

        if res.status_code == 401:  # uh oh!
            if reauthorize:
                self.iam_client.clear_token()
                return self.create_dataset(
                    project_uuid=project_uuid,
                    path=path,
                    description=description,
                    data=data,
                    data_type=data_type,
                    reauthorize=False,
                )
            raise UrsaNotAuthorized("forbidden")
        if res.status_code == 409:
            return {"accepted": False, "path": path}
        if res.status_code >= 500:
            print("Unable to create dataset. Internal platform error.")
            return {"accepted": False, "path": path}

        dataset_id = res.headers["x-dataset-id"]
        # if successful, we get the dataset id from the response headers
        # if we have a description, we use GraphQL to set the description

        if description is not None:
            variables = {
                "id": dataset_id,
                "input": {"description": description, "isSynced": False},
            }

            query_response = self.raw_query(
                query=UPDATE_DATASET_MUTATION, variables=variables
            )
            if "errors" in query_response["data"]["updateDataset"]:
                return {"accepted": False, "path": path, "id": dataset_id}

        return {"accepted": True, "path": path, "id": dataset_id}

    def sync_dataset(
        self,
        project_uuid=None,
        path=None,
        description=None,
        data=None,
        data_type=None,
        reauthorize=True,
    ):
        """
        Syncs a dataset, returning the uuid and id of the dataset.

        If no data is provided, the dataset is created with no attached data.
        Data can be attached later.
        """
        # we first use the graphql to create the dataset
        # then we use a PUT to upload the data
        # the Python graphql client doesn't support multi-part HTTP requests
        self._raise_if_none(project_uuid, "project_uuid")
        self._raise_if_none(path, "path")
        self._raise_if_none(data, "data")
        self._raise_if_none(data_type, "data_type")

        res = requests.head(
            url=f"{self.endpoint}api/dss/{project_uuid}/{path}",
            headers={
                "Authorization": f"Bearer {self.iam_client.get_token()}",
                "X-UF-Sync": "true",
            },
        )
        if res.status_code == 401:  # uh oh!
            if reauthorize:
                self.iam_client.clear_token()
                return self.sync_dataset(
                    project_uuid=project_uuid,
                    path=path,
                    description=description,
                    data=data,
                    data_type=data_type,
                    reauthorize=False,
                )
            raise UrsaNotAuthorized("forbidden")
        if res.status_code == 404 or not self._file_is_same(data, res):
            # if we get here, we need to create/update the dataset
            # first, we POST to the right URL to upload the contents of the file
            with open(data, "rb") as file:
                contents = file.read()
            res = requests.post(
                url=f"{self.endpoint}api/dss/{project_uuid}/{path}",
                data=contents,
                headers={
                    "Content-Type": data_type,
                    "Authorization": f"Bearer {self.iam_client.get_token()}",
                    "X-UF-Sync": "true",
                },
            )

        if res.status_code == 401:  # uh oh!
            raise UrsaNotAuthorized("forbidden")
        if res.status_code == 409:
            return {"accepted": False, "path": path}
        if res.status_code >= 500:
            print("Unable to create dataset. Internal platform error.")
            return {"accepted": False, "path": path}

        dataset_id = res.headers["x-dataset-id"]
        # if successful, we get the dataset id from the response headers
        # if we have a description, we use GraphQL to set the description

        if description is not None:
            variables = {
                "id": dataset_id,
                "input": {"description": description, "isSynced": True},
            }

            query_response = self.raw_query(
                query=UPDATE_DATASET_MUTATION, variables=variables
            )
            if "errors" in query_response["data"]["updateDataset"]:
                return {"accepted": False, "path": path, "id": dataset_id}

        return {"accepted": True, "path": path, "id": dataset_id}

    def get_dataset_url(self, project_uuid=None, dataset=None):
        if self.is_uuid(dataset):
            variables = {"id": dataset}
            query_response = self.raw_query(
                query=GET_DATASET_PATH_QUERY, variables=variables
            )
            if "errors" in query_response:
                return None
            path = query_response["data"]["dataset"]["path"]
            # handle_name = query_response['data']['dataset']['project']['handleName']
            owner_handle = query_response["data"]["dataset"]["project"]["user"][
                "handle"
            ]
            project_name = query_response["data"]["dataset"]["project"]["name"]
            handle_name = f"{owner_handle}/{project_name}"
            return f"{self.endpoint}api/dss/{urllib.parse.quote(handle_name)}/{urllib.parse.quote(path)}"
        else:
            return f"{self.endpoint}api/dss/{urllib.parse.quote(project_uuid)}/{urllib.parse.quote(dataset)}"

    def get_dataset_details(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {"id": dataset_id}
        else:
            dataset = self.get_dataset(dataset_id, project_uuid=project_uuid)
            if dataset is None:
                return None
            variables = {"id": dataset["id"]}
        query_response = self.raw_query(query=GET_DATASET_QUERY, variables=variables)
        if "errors" in query_response:
            return None
        return query_response["data"]["dataset"]

    def get_dataset(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {"id": dataset_id}
            query_response = self.raw_query(
                query=GET_DATASET_QUERY, variables=variables
            )
            return query_response["data"]["dataset"]
        else:
            variables = {"projectId": project_uuid, "path": dataset_id}
            query_response = self.raw_query(
                query=GET_DATASET_BY_PATH_QUERY, variables=variables
            )
            return query_response["data"]["project"]["datasets"][0]

    def download_dataset(self, url, local_path, reauthorize=True):
        with open(local_path, "wb") as fd:
            r = requests.get(
                url,
                stream=True,
                headers={"authorization": f"Bearer {self.iam_client.get_token()}"},
            )
            if r.status_code >= 400:
                if reauthorize:
                    self.iam_client.clear_token()
                    return self.download_dataset(url, local_path, reauthorize=False)
                return False
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
        return True

    def download_dataset_content(self, url, reauthorize=True):
        local = bytearray()
        r = requests.get(
            url,
            stream=True,
            headers={"authorization": f"Bearer {self.iam_client.get_token()}"},
        )
        if r.status_code >= 400:
            if reauthorize:
                self.iam_client.clear_token()
                return self.download_dataset_content(url, reauthorize=False)
            return False
        for chunk in r.iter_content(chunk_size=1024 * 1024):
            local += chunk
        content = local.decode("utf-8")
        return content

    def delete_dataset(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {"id": dataset_id}
        else:
            dataset = self.get_dataset(dataset_id, project_uuid=project_uuid)
            if dataset is None:
                return False
            variables = {"id": dataset["id"]}
        query_response = self.raw_query(
            query=DELETE_DATASET_MUTATION, variables=variables
        )
        if "errors" in query_response and query_response["errors"]:
            return False
        return True

    def list_datasets(self, project_scope=None):
        if project_scope is None:
            raise ValueError("project_scope must be provided")
        variables = {}
        query = LIST_DATASETS_FOR_PROJECT_QUERY
        variables = {"projectId": project_scope}
        query_response = self.raw_query(query=query, variables=variables)

        if "errors" in query_response:
            return []
        if query_response["data"]["project"] is None:
            raise UrsaBadProjectName(project_scope)
        return query_response["data"]["project"]["datasets"]

    def _file_is_same(self, local_path, head_resp):
        server_md5 = head_resp.headers["x-uf-md5-hash"]
        server_sha256 = head_resp.headers["x-uf-sha256-hash"]
        server_sha1 = head_resp.headers["x-uf-sha1-hash"]

        with open(local_path, "rb") as file:
            contents = file.read()
            client_md5 = hashlib.md5(contents).hexdigest()
            client_sha256 = hashlib.sha256(contents).hexdigest()
            client_sha1 = hashlib.sha1(contents).hexdigest()

        return (
            client_md5 == server_md5
            and client_sha256 == server_sha256
            and client_sha1 == server_sha1
        )



import asyncio


class IncomingMessage:
    def __init__(self, _join_ref, msg_ref, topic, event, payload):
        self.event = event
        self.topic = topic
        self.payload = payload
        self.ref = msg_ref

    def reply(self, event, payload):
        return OutgoingMessage(None, self.ref, self.topic, event, payload)


class OutgoingMessage:
    def __init__(self, join_ref, msg_ref, topic, event, payload):
        self.event = event
        self.topic = topic
        self.payload = payload
        self.ref = msg_ref
        self.join_ref = join_ref
        self.sent = asyncio.Future()

    def send(self, socket):
        socket.send(self)
        return self.sent



"""
Exception classes for Ursactl
"""


class UrsaCtlError(Exception):
    """Generic errors."""


class UrsaNotAuthorized(UrsaCtlError):
    """Authentication/authorization errors."""


class UrsaProjectNotDefined(UrsaCtlError):
    """Project required but not defined."""


class UrsaBadProjectName(UrsaCtlError):
    """Project name is not valid."""



"""
Planning capabilities client.
"""

import json

from ursactl.core.exc import UrsaBadProjectName

from ._base import Base

LIST_PACKAGES_QUERY = """\
query list_packages($id: String!) {
    project(handleName: $id) {
        packages {
            id
            name
            isSynced
        }
    }
}
"""

LIST_AGENTS_QUERY = """\
query list_packages($id: String!) {
    project(handleName: $id) {
        agents {
            id
            name
            isSynced
        }
    }
}
"""

LIST_AGENT_SWARMS_QUERY = """\
query list_packages($id: String!) {
    project(handleName: $id) {
        agent_swarms {
            id
            name
        }
    }
}
"""

CREATE_PACKAGE_MUTATION = """\
mutation create_package($input: CreatePackageInput!) {
    createPackage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

SYNC_PACKAGE_MUTATION = """\
mutation sync_package($input: SyncPackageInput!) {
    syncPackage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

UPDATE_PACKAGE_MUTATION = """\
mutation update_package($input: UpdatePackageInput!) {
    updatePackage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

GET_PACKAGE_BY_NAME = """\
query get_package($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        packages(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
        }
    }
}
"""

GET_PACKAGE_BY_ID = """\
query get_package($id: ID!) {
    package(id: $id) {
        id
        name
    }
}
"""

DELETE_PACKAGE_MUTATION = """\
mutation delete_package($id: ID!) {
    deletePackage(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

DELETE_SYNCED_PACKAGE_MUTATION = """\
mutation delete_synced_package($id: ID!) {
    deleteSyncedPackage(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

CREATE_AGENT_MUTATION = """\
mutation create_agent($input: CreateAgentInput!) {
    createAgent(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

SYNC_AGENT_MUTATION = """\
mutation sync_agent($input: SyncAgentInput!) {
    syncAgent(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

DELETE_AGENT_MUTATION = """\
mutation delete_agent($id: ID!) {
    deleteAgent(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

DELETE_SYNCED_AGENT_MUTATION = """\
mutation delete_synced_agent($id: ID!) {
    deleteSyncedAgent(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

GET_AGENT_BY_NAME = """\
query get_agent($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        agents(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
        }
    }
}
"""

GET_AGENT_BY_ID = """\
query get_agent($id: ID!) {
    agent(id: $id) {
        id
        name
    }
}
"""

GET_AGENT_SWARM_BY_NAME = """\
query get_agent_swarm($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        agentSwarms(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
        }
    }
}
"""

GET_AGENT_SWARM_BY_ID = """\
query get_agent_swarm($id: ID!) {
    agentSwarm(id: $id) {
        id
        name
    }
}
"""

RUN_AGENT_MUTATION = """\
mutation run_agent($input: RunAgentInput!) {
    runAgent(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
        }
    }
}
"""

STOP_AGENT_MUTATION = """\
mutation stop_agent($id: ID!) {
    stopAgent(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
        }
    }
}
"""

CREATE_AGENT_SWARM_MUTATION = """\
mutation create_agent_swarm($input: CreateAgentSwarmInput!) {
    createAgentSwarm(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
        }
    }
}
"""

DELETE_AGENT_SWARM_MUTATION = """\
mutation delete_agent_swarm($id: ID!) {
    deleteAgentSwarm(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

RUN_AGENTS_IN_SWARM_MUTATION = """\
mutation run_agents_in_swarm($agentSwarmId: ID!, $agent: String!, $groups: [String], $params: [AgentState]) {
    runAgentsInSwarm(agent: $agent, agentSwarmId: $agentSwarmId, groups: $groups, params: $params) {
        errors {
            code
            message
            fields
        }
        result
    }
}
"""

LIST_RUNNING_AGENTS_BY_PROJECT = """\
query list_agents_by_project($id: String!) {
    project(handleName: $id) {
        agents {
            name
            runningAgents {
                id
                startedAt
                status
            }
        }
    }
}
"""

LIST_RUNNING_AGENTS_BY_AGENT = """\
query list_agents_by_agent($id: ID!) {
    agent(id: $id) {
        name
        runningAgents {
            id
            startedAt
            status
        }
    }
}
"""

SEND_EVENT_TO_AGENT_MUTATION = """\
mutation send_event($agentId: ID!, $domain: String!, $name: String!, $params: Json!) {
  sendEventToAgent(agentId:$agentId,domain:$domain,name:$name,params:$params) {
    errors {
        code
        message
        fields
    }
    result {
        domain
        name
        params
    }
  }
}
"""

TEST_EVENT_WITH_AGENT_MUTATION = """\
mutation test_event($agentId: ID!, $domain: String!, $name: String!, $params: Json) {
    testEventWithAgent(agentId:$agentId,domain:$domain,name:$name,params:$params) {
        errors {
            code
            message
            fields
        }
        result
    }
}
"""

SEND_EVENT_TO_AGENT_SWARM_MUTATION = """\
        mutation send_event($agentSwarmId: ID!, $domain: String!, $name: String!, $params: Json!, $group: String) {
  sendEventToAgentSwarm(agentSwarmId:$agentSwarmId,domain:$domain,name:$name,params:$params,group:$group) {
    errors {
        code
        message
        fields
    }
    result {
        domain
        name
        params
    }
  }
}
"""


class PlanningClient(Base):
    """-
    Client providing access to the planning capabilities.
    """

    def list_packages(self, project_uuid=None):
        """
        Return a list of available packages.
        """

        query_response = self.raw_query(
            query=LIST_PACKAGES_QUERY, variables={"id": project_uuid}
        )
        if query_response["data"]["project"] is None:
            raise UrsaBadProjectName(project_uuid)
        return query_response["data"]["project"]["packages"]

    def list_agents(self, project_uuid=None):
        """
        Return a list of available agents.
        """
        query_response = self.raw_query(
            query=LIST_AGENTS_QUERY, variables={"id": project_uuid}
        )
        if query_response["data"]["project"] is None:
            raise UrsaBadProjectName(project_uuid)
        return query_response["data"]["project"]["agents"]

    def list_running_agents(self, agent_id=None, project_uuid=None):
        """
        Return a list of running agents.
        """
        if agent_id is not None:
            if self.is_uuid(agent_id):
                variables = {"id": agent_id}
            else:
                agent = self.get_agent(agent_id, project_uuid)
                if agent is None:
                    return []
                variables = {"id": agent["id"]}
            query_response = self.raw_query(
                query=LIST_RUNNING_AGENTS_BY_AGENT, variables=variables
            )
            agent_name = query_response["data"]["agent"]["name"]
            return [
                {"agent": agent_name, **info}
                for info in query_response["data"]["agent"]["runningAgents"]
            ]

        elif project_uuid is not None:
            query_response = self.raw_query(
                query=LIST_RUNNING_AGENTS_BY_PROJECT, variables={"id": project_uuid}
            )
            if query_response["data"]["project"] is None:
                raise UrsaBadProjectName(project_uuid)
            return [
                {"agent": agent["name"], **info}
                for agent in query_response["data"]["project"]["agents"]
                for info in agent["runningAgents"]
            ]

    def list_agent_swarms(self, project_uuid=None):
        """
        Return a list of available agents.
        """
        query_response = self.raw_query(
            query=LIST_AGENT_SWARMS_QUERY, variables={"id": project_uuid}
        )
        if query_response["data"]["project"] is None:
            raise UrsaBadProjectName(project_uuid)
        return query_response["data"]["project"]["agent_swarms"]

    def create_package(self, project_uuid=None, content=None):
        """
        Create a new package.
        """
        mutation_response = self.raw_query(
            query=CREATE_PACKAGE_MUTATION,
            variables={"input": {"source": content, "project": project_uuid}},
        )
        return mutation_response["data"]["createPackage"]

    def sync_package(self, project_uuid=None, content=None):
        """
        Sync a new package.
        """
        mutation_response = self.raw_query(
            query=SYNC_PACKAGE_MUTATION,
            variables={"input": {"source": content, "project": project_uuid}},
        )
        return mutation_response["data"]["syncPackage"]

    def update_package(self, project_uuid=None, content=None):
        """
        Update an existing package.
        """
        mutation_response = self.raw_query(
            query=UPDATE_PACKAGE_MUTATION,
            variables={"input": {"source": content, "project": project_uuid}},
        )
        return mutation_response["data"]["updatePackage"]

    def create_agent(self, project_uuid=None, name=None, packages=None, behaviors=None):
        """
        Create a new agent.
        """
        mutation_response = self.raw_query(
            query=CREATE_AGENT_MUTATION,
            variables={
                "input": {
                    "name": name,
                    "packages": packages,
                    "behaviors": behaviors,
                    "project": project_uuid,
                }
            },
        )
        return mutation_response["data"]["createAgent"]

    def sync_agent(self, project_uuid=None, name=None, packages=None, behaviors=None):
        """
        Sync an agent.
        """
        mutation_response = self.raw_query(
            query=SYNC_AGENT_MUTATION,
            variables={
                "input": {
                    "name": name,
                    "packages": packages,
                    "behaviors": behaviors,
                    "project": project_uuid,
                }
            },
        )
        return mutation_response["data"]["syncAgent"]

    def create_agent_swarm(self, project_uuid=None, name=None):
        """
        Create a new swarm.
        """
        mutation_response = self.raw_query(
            query=CREATE_AGENT_SWARM_MUTATION,
            variables={"input": {"name": name, "project": project_uuid}},
        )
        return mutation_response["data"]["createAgentSwarm"]

    def get_package(self, name=None, project_uuid=None):
        """
        Get the id and basic info about a package.
        """
        if self.is_uuid(name):
            variables = {"id": name}
            query = GET_PACKAGE_BY_ID
        else:
            variables = {"name": name, "project_id": project_uuid}
            query = GET_PACKAGE_BY_NAME
        result = self.raw_query(query=query, variables=variables)
        if query == GET_PACKAGE_BY_ID:
            return result["data"]["package"]
        return result["data"]["project"]["packages"][0]

    def get_agent(self, name=None, project_uuid=None):
        """
        Get the id and basic info about an agent.
        """
        if self.is_uuid(name):
            variables = {"id": name}
            query = GET_AGENT_BY_ID
        else:
            variables = {"name": name, "project_id": project_uuid}
            query = GET_AGENT_BY_NAME
        result = self.raw_query(query=query, variables=variables)
        if query == GET_AGENT_BY_ID:
            return result["data"]["agent"]

        return result["data"]["project"]["agents"][0]

    def get_agent_swarm(self, name=None, project_uuid=None):
        """
        Get the id and basic info about an agent swarm.
        """
        if self.is_uuid(name):
            variables = {"id": name}
            query = GET_AGENT_SWARM_BY_ID
        else:
            variables = {"name": name, "project_id": project_uuid}
            query = GET_AGENT_SWARM_BY_NAME
        result = self.raw_query(query=query, variables=variables)
        if query == GET_AGENT_SWARM_BY_ID:
            return result["data"]["agentSwarm"]

        return result["data"]["project"]["agentSwarms"][0]

    def delete_package(self, package_id, project_uuid=None):
        """
        Deletes a package.
        """
        if self.is_uuid(package_id):
            variables = {"id": package_id}
        else:
            package = self.get_package(name=package_id, project_uuid=project_uuid)
            if package is None:
                return {}
            variables = {"id": package["id"]}

        result = self.raw_query(query=DELETE_PACKAGE_MUTATION, variables=variables)

        return result["data"]["deletePackage"]

    def delete_synced_package(self, package_id, project_uuid=None):
        """
        Deletes a synced package.
        """
        if self.is_uuid(package_id):
            variables = {"id": package_id}
        else:
            package = self.get_package(name=package_id, project_uuid=project_uuid)
            if package is None:
                return {}
            variables = {"id": package["id"]}

        result = self.raw_query(
            query=DELETE_SYNCED_PACKAGE_MUTATION, variables=variables
        )

        return result["data"]["deleteSyncedPackage"]

    def delete_agent(self, agent_id, project_uuid=None):
        """
        Deletes an agent.
        """
        if self.is_uuid(agent_id):
            variables = {"id": agent_id}
        else:
            agent = self.get_agent(name=agent_id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables = {"id": agent["id"]}

        result = self.raw_query(query=DELETE_AGENT_MUTATION, variables=variables)

        return result["data"]["deleteAgent"]

    def delete_agent_swarm(self, agent_swarm_id, project_uuid=None):
        """
        Deletes an agent swarm.
        """
        if self.is_uuid(agent_swarm_id):
            variables = {"id": agent_swarm_id}
        else:
            agent = self.get_agent_swarm(name=agent_swarm_id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables = {"id": agent["id"]}

        result = self.raw_query(query=DELETE_AGENT_SWARM_MUTATION, variables=variables)

        return result["data"]["deleteAgentSwarm"]

    def delete_synced_agent(self, agent_id, project_uuid=None):
        """
        Deletes a synced agent.
        """
        if self.is_uuid(agent_id):
            variables = {"id": agent_id}
        else:
            agent = self.get_agent(name=agent_id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables = {"id": agent["id"]}

        result = self.raw_query(query=DELETE_SYNCED_AGENT_MUTATION, variables=variables)

        return result["data"]["deleteSyncedAgent"]

    def run_agent(
        self, agent_id, project_uuid=None, initial_predicates=None, initial_state=None
    ):
        """
        Runs an agent.
        """
        variables = {"input": {"project": project_uuid}}
        if self.is_uuid(agent_id):
            agent = self.get_agent(name=agent_id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables["input"]["agent"] = agent["name"]
        else:
            agent = self.get_agent(name=agent_id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables["input"]["agent"] = agent_id

        if initial_state is not None:
            variables["input"]["initialState"] = initial_state
        if initial_predicates is not None:
            variables["input"]["initialPredicates"] = initial_predicates

        result = self.raw_query(query=RUN_AGENT_MUTATION, variables=variables)

        return result["data"]["runAgent"]

    def run_agents_in_swarm(self, agent_swarm_id, agent_name, configs, groups=[]):
        """
        Runs agents in an agent swarm.
        """
        variables = {
            "agent": agent_name,
            "agentSwarmId": agent_swarm_id,
            "params": configs,
        }
        if groups:
            variables["groups"] = groups
        result = self.raw_query(query=RUN_AGENTS_IN_SWARM_MUTATION, variables=variables)

        return result["data"]["runAgentsInSwarm"]

    def send_event_to_agent(self, agent_instance_id, event):
        """
        Sends an event to an agent.
        """
        variables = {"agentId": agent_instance_id, **event}
        variables["params"] = json.dumps(variables["params"])

        result = self.raw_query(query=SEND_EVENT_TO_AGENT_MUTATION, variables=variables)

        return result["data"]["sendEventToAgent"]

    def test_event_with_agent(self, agent_instance_id, event):
        """
        Sends an event to an agent to test planning.
        """
        variables = {"agentId": agent_instance_id, **event}
        variables["params"] = json.dumps(variables["params"])

        result = self.raw_query(query=TEST_EVENT_WITH_AGENT_MUTATION, variables=variables)

        return result["data"]["testEventWithAgent"]

    def stop_agent(self, agent_id):
        """
        Stops an agent.
        """
        variables = {"id": agent_id}

        result = self.raw_query(query=STOP_AGENT_MUTATION, variables=variables)

        return result["data"]["stopAgent"]

    def send_event_to_agent_swarm(self, agent_swarm_id, event):
        """
        Sends an event to an agent swarm.
        """
        variables = {"agentSwarmId": agent_swarm_id, **event}
        variables["params"] = json.dumps(variables["params"])

        result = self.raw_query(
            query=SEND_EVENT_TO_AGENT_SWARM_MUTATION, variables=variables
        )

        return result["data"]["sendEventToAgentSwarm"]



import asyncio

from .transport import Transport

# from .messages import IncomingMessage, OutgoingMessage
from .utils import DONE, get_unless_done


class Socket:
    def __init__(self, url, params, pending_queue):
        self.url = url
        self.params = params
        self._incoming_queue = asyncio.Queue()
        self._outgoing_queue = asyncio.Queue()
        self._pending_queue = pending_queue
        self.connected = False
        self._response_futures = {}

    async def connect(self):
        if self.connected:
            raise Exception("Already connected")

        self.transport = Transport(
            self.url, self.params, self._incoming_queue, self._outgoing_queue
        )
        transport_task = asyncio.ensure_future(self.transport.run())
        await self.transport.ready

        self._transport_task = transport_task
        self._done_recv = asyncio.Future()
        self._recv_task = asyncio.ensure_future(self._recv_loop())
        self.connected = True

    async def disconnect(self):
        if self.connected:
            self._done_recv.set_result(True)

            await asyncio.gather(self._recv_task, self._transport_task)
            await self.transport.stop()

            self.connected = False

    async def stop(self):
        await self.disconnect()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def _recv_loop(self):
        while True:
            message = await get_unless_done(self._incoming_queue.get(), self._done_recv)
            if message is DONE:
                return

            if message.ref in self._response_futures:
                future = self._response_futures.pop(message.ref)
                future.set_result(message.payload)
            else:
                self._pending_queue.put_nowait(message)

    def send(self, message):
        if not self.connected:
            raise Exception("Not connected")
        self._outgoing_queue.put_nowait(message)
        return message.sent



from cement import Controller, ex

from ..core.services import client


class Get(Controller):
    """
    Provides the 'get' verb.
    """

    class Meta:
        label = "get"
        stacked_on = "base"
        stacked_type = "nested"
        help = "get something"

    @ex(
        help="download dataset",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (
                ["dataset"],
                {"help": "UUID or path of the dataset to download", "action": "store"},
            ),
            (
                ["destination"],
                {"help": "local file in which the dataset is saved", "action": "store"},
            ),
        ],
    )
    def dataset(self):
        dataset_uuid = self.app.pargs.dataset
        destination = self.app.pargs.destination
        dss_client = client("dss", self.app)
        url = dss_client.get_dataset_url(
            project_uuid=self._project_scope, dataset=dataset_uuid
        )
        if url is None:
            exit(1)
        if not dss_client.download_dataset(url, destination):
            exit(1)

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")



import time

from requests.exceptions import ConnectionError

from ursactl.core._base import Base
from ursactl.core.services import client

from .pipeline_run import PipelineRun


class PipelineSweep(Base):
    """
    Provides access to a pipeline sweep.
    """

    def __init__(self, *args, pipeline=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = pipeline

    @property
    def client(self):
        if self._client is None:
            self._client = client("ape", self.app)
        return self._client

    @property
    def pipeline_runs(self):
        return [
            PipelineRun(
                uuid=run["id"],
                project=self.project,
                app=self.app,
                client=self.client,
                pipeline=self.pipeline,
                **run
            )
            for run in self._data["pipelineRuns"]
        ]

    @property
    def job_status(self):
        return self._data["jobStatus"]

    def refresh(self):
        if self.uuid is not None:
            if self._cached_data is None:
                self._cached_data = {}
            self._cached_data.update(self.client.get_pipeline_sweep(uuid=self.uuid))

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    "pipelineRuns": [],
                    "jobStatus": None,
                }
            else:
                self._cached_data = self.client.get_pipeline_sweep(uuid=self.uuid)
        return self._cached_data

    def __enter__(self):
        while not self.completed:
            time.sleep(5)
            try:
                self.refresh()
            except ConnectionError:
                pass
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @property
    def completed(self):
        return self.job_status == "completed" and all(
            run.completed for run in self.pipeline_runs
        )



import sys
from pathlib import Path

from cement import Controller


class Init(Controller):
    """
    Provides the 'init' verb
    """

    class Meta:
        label = "init"
        stacked_on = "base"
        stacked_type = "nested"
        help = "initialize a directory for sync with Ursa Frontier"
        arguments = [
            (
                ["--dir", "--directory"],
                {
                    "help": "directory to initialize (defaults to 'ursa')",
                    "dest": "directory",
                    "action": "store",
                    "default": "ursa",
                },
            ),
        ]

    def _default(self):
        root = Path(self.app.pargs.directory)
        if not root.exists():
            root.mkdir(parents=True)
        elif not root.is_dir():
            print("Error: %s exists and is not a directory." % root)
            sys.exit(1)

        for subdir in (
            "transforms",
            "generators",
            "pipelines",
            "datasets",
            "packages",
            "agents",
            "pages",
        ):
            (root / subdir).mkdir(parents=True, exist_ok=True)
            (root / subdir / ".keep").touch()



"""
Asynchronous client for the websocket interface.
"""

import asyncio

from .messages import OutgoingMessage
from .socket import Socket
from .utils import get_unless_done


class AsyncClient:
    def __init__(self, endpoint: str, app):
        from .. import client

        self.app = app
        self.endpoint = "ws" + endpoint[4:]
        self.iam_client = client("iam", app)
        self._ref = 1
        self._pending_queue = asyncio.Queue()

    async def connect(self):
        params = {"token": self.iam_client.get_token()}

        self.socket = Socket(
            f"{self.endpoint}/api/socket/websocket", params, self._pending_queue
        )
        self._done_recv = asyncio.Future()
        await self.socket.connect()

    async def disconnect(self):
        await self.socket.disconnect()

    async def stop(self):
        self._done_recv.set_result(True)
        await self.socket.stop()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def join(self, topic):
        message = OutgoingMessage(self._ref, None, topic, "phx_join", {})
        self._ref += 1
        await message.send(self.socket)

    async def send(self, event, topic, payload, ref=None):
        if ref is None:
            ref = self._ref
            self._ref += 1
        message = OutgoingMessage(None, ref, topic, event, payload)
        await message.send(self.socket)

    async def receive(self):
        msg = await get_unless_done(self._pending_queue.get(), self._done_recv)

        return msg



import json

import yaml
from cement import Controller, ex

from ..core.services import client


class Run(Controller):
    """
    Provides the 'run' verb.
    """

    class Meta:
        label = "run"
        stacked_on = "base"
        stacked_type = "nested"
        help = "run something"

    @ex(
        help="run agent",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["agent"], {"help": "agent uuid or path", "action": "store"}),
        ],
    )
    def agent(self):
        planning_client = client("planning", self.app)
        result = planning_client.run_agent(
            self.app.pargs.agent, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="run pipeline",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (
                ["--param"],
                {
                    "help": "pipeline parameter",
                    "dest": "parameters",
                    "action": "append",
                },
            ),
            (
                ["--sweep-param"],
                {
                    "help": "pipeline sweep parameter",
                    "dest": "sweep_parameters",
                    "action": "append",
                },
            ),
            (["pipeline"], {"help": "pipeline path", "action": "store"}),
        ],
    )
    def pipeline(self):
        ape_client = client("ape", self.app)
        pipeline = self.app.pargs.pipeline
        parameters = None
        sweep_parameters = None
        if self.app.pargs.parameters:
            parameters = {
                k: json.loads(v)
                for k, v in dict(
                    item.split("=") for item in self.app.pargs.parameters
                ).items()
            }
        if self.app.pargs.sweep_parameters:
            sweep_parameters = {
                k: json.loads(v)
                for k, v in dict(
                    item.split("=") for item in self.app.pargs.sweep_parameters
                ).items()
            }

        result = ape_client.run_pipeline(
            pipeline=pipeline,
            parameters=parameters,
            sweep_parameters=sweep_parameters,
            project=self._project_scope,
        )
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")



import asyncio

from ursactl.core._base import Base
from ursactl.core.services import client

from .running_agent import RunningAgent
from .services.async_client.utils import DONE


class AgentSwarm(Base):
    """
    Provides access to agent swarms.
    """

    @property
    def client(self):
        if self._client is None:
            self._client = client("planning", self.app)
        return self._client

    @property
    def name(self):
        return self._data["name"]

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {"name": None, "project_uuid": None}
            else:
                self._cached_data = self.client.get_agent_swarm(self.uuid)
        return self._cached_data

    def send_event(self, domain, name, params, group=None):
        result = self.client.send_event_to_agent_swarm(
            self.uuid,
            {"domain": domain, "name": name, "params": params, "group": group},
        )
        return result["result"]

    def send_events(self, events):
        actions = []
        for event in events:
            result = self.client.send_event_to_agent_swarm(self.uuid, event)
            if result["result"]:
                actions.extend(result["result"])
        return actions

    def add_agents(self, agent_name, configs, groups=[]):
        result = self.client.run_agents_in_swarm(self.uuid, agent_name, configs, groups)
        if result["errors"]:
            return None
        return [
            RunningAgent(
                uuid=running_agent_id,
                app=self.app,
                client=self.client,
                project=self.project,
            )
            for running_agent_id in result["result"]
        ]

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.client.delete_agent_swarm(self.uuid)

    def for_async(self):
        return AsyncAgentSwarm(self)

    @classmethod
    def create_anonymous(klass, project=None, app=None, **_kwargs):
        c = client("planning", app)
        swarm = c.create_agent_swarm(project_uuid=project.uuid)
        return klass(uuid=swarm["result"]["id"], project=project, app=app, client=c)


class AsyncAgentSwarm(Base):
    def __init__(self, swarm):
        self.swarm = swarm
        self._app = swarm.app
        self._uuid = swarm.uuid
        self._client = None
        self._topic = f"swarm:{self._uuid}"
        self.project = swarm.project
        self._handlers = {}

    @property
    def client(self):
        if self._client is None:
            self._client = client("async", self.app)
        return self._client

    async def stop(self):
        await self.client.stop()
        return True

    def add_handler(self, domain, handler):
        print("Adding handler", domain, handler)
        try:
            self._handlers[domain].append(handler)
        except KeyError:
            self._handlers[domain] = [handler]
        print(self._handlers)

    async def connect(self):
        await self.client.connect()
        await self.client.join(self._topic)
        recv_task = asyncio.ensure_future(self.run())
        self._recv_task = recv_task

    async def disconnect(self):
        await asyncio.gather(self._recv_task)
        await self.client.disconnect()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def send_event(self, domain, name, params, group=None):
        print("Sending event", domain, name, params, group)
        await self.client.send(
            "event",
            self._topic,
            {"domain": domain, "name": name, "params": params, "group": group},
        )

    async def receive(self):
        return await self.client.receive()

    def add_agents(self, agent_name, configs, groups=[]):
        return self.swarm.add_agents(agent_name, configs, groups)

    async def _handle_action(self, message):
        payload = message.payload
        if payload["domain"] in self._handlers:
            handlers = self._handlers[payload["domain"]]
            for handler in handlers:
                try:
                    method = handler.__class__.actions.get(payload["name"])
                    events = await method(handler, payload["params"])
                    if events:
                        for event in events:
                            await self.send_event(
                                event.get("domain"),
                                event.get("name"),
                                event.get("params"),
                                event.get("group"),
                            )
                except Exception as e:
                    print(e)
                    pass

    async def _hande_perception(self, message):
        payload = message.payload
        handled = False
        if payload["domain"] in self._handlers:
            handlers = self._handlers[payload["domain"]]
            for handler in handlers:
                try:
                    method = handler.__class__.perceptions.get(payload["name"])
                    value = await method(handler, *payload["arguments"])
                    await self.client.send(
                        "perceived",
                        self._topic,
                        {
                            "domain": payload["domain"],
                            "name": payload["name"],
                            "arguments": payload["arguments"],
                            "value": value,
                        },
                    )

                    handled = True
                    break
                except Exception as e:
                    print(e)
                    pass
        if not handled:
            await self.client.send(
                "perceived",
                self._topic,
                {
                    "domain": payload["domain"],
                    "name": payload["name"],
                    "arguments": payload["arguments"],
                    "value": None,
                },
            )

    async def run(self):
        while True:
            message = await self.client.receive()
            if message == DONE:
                return
            if message.event.startswith("phx_"):
                print("Received phoenix message", message.event, message.payload)
                continue
            if message.event == "action":
                await self._handle_action(message)
                continue
            if message.event == "perceive":
                await self._hande_perception(message)
                continue

            print(f"Received message with unknown event: {message.event}")



"""
Utilities for working with packages in the CLI.
"""

import re


def load_package(file):
    loader = load_source
    if file.suffix == ".md":
        loader = load_from_markdown

    with open(file, "r", encoding="utf-8") as fd:
        source = loader(fd)
        info = {"name": None, "file": file, "kind": "package"}
        for line in source:
            match = re.fullmatch(r"\s*package\s+(\S+)\s*([;#].*)?", line)
            if match:
                if info["name"] is not None:
                    raise ValueError(f"Multiple packages found in {file}")
                info["name"] = match.group(1)

        info["content"] = "".join(source)

    return info


def load_source(fd):
    return list(fd)


def load_from_markdown(fd):
    source = list(fd)
    if re.match(r"^---\s*\n", source[0]):
        split_at = len(source)
        for i in range(1, len(source) - 1):
            if re.match(r"^---\s*\n", source[i]):
                split_at = i
                break
        source = source[split_at + 1 :]

    buffer = []
    in_fenced_area = False
    for line in source:
        if not in_fenced_area and re.match(r"^```(htn)?\n$", line):
            in_fenced_area = True
        elif in_fenced_area and re.match(r"^```\n$", line):
            in_fenced_area = False
        elif in_fenced_area:
            buffer.append(line)
    return buffer



"""
Utilities for loading pages and getting metadata from the header.
"""

import sys

import yaml


def load_page(file):
    if file.endswith(".md"):
        loader = load_from_markdown
    else:
        print("Only Markdown files may be uploaded.")
        sys.exit(1)

    with open(file, "r", encoding="utf-8") as fd:
        pages = loader(fd)
    return pages


def load_from_markdown(source):
    info = {}
    content = list(source)
    if content[0] == "---\n":
        split_at = len(content)
        # get metadata from header
        for i in range(1, len(content) - 1):
            if content[i] == "---\n":
                split_at = i
                break
        raw_yaml = "".join(content[1:split_at])
        content = content[split_at + 1 :]
        info = yaml.load(raw_yaml, Loader=yaml.Loader)

    title, content = _get_title(content)

    if title is not None:
        info["title"] = title
    info["content"] = "".join(content)
    info["kind"] = "page"

    return info


def _get_title(content):
    title = None
    # the first top-level header is the name of the transform
    for i in range(0, len(content) - 1):
        if content[i].startswith("# "):
            title = content[i][2:].strip()
            content = content[i + 1 :]
            break
    return title, content



import json
import sys

import yaml
from cement import Controller, ex

from ursactl.controllers.utils.transforms import load_transform

from ..core.services import client


def _yaml_load(fd):
    return list(yaml.load_all(fd, Loader=yaml.Loader))


class Update(Controller):
    """
    Provides the 'update' verb.
    """

    class Meta:
        label = "update"
        stacked_on = "base"
        stacked_type = "nested"
        help = "update something"

    @ex(
        help="update an existing project",
        arguments=[
            (["id"], {"help": "project identifier", "action": "store"}),
            (["--name"], {"help": "project name", "action": "store", "dest": "name"}),
            (
                ["--description"],
                {
                    "help": "project description",
                    "action": "store",
                    "dest": "description",
                },
            ),
        ],
    )
    def project(self):
        projects_client = client("projects", self.app)
        result = projects_client.update(
            id=self.app.pargs.id,
            name=self.app.pargs.name,
            description=self.app.pargs.description,
        )
        self._print(result)

    @ex(
        help="update an existing transform",
        arguments=[
            (["file"], {"help": "transform definition file", "action": "store"}),
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def transform(self):
        project_uuid = self._project_scope
        ape_client = client("ape", self.app)

        file = self.app.pargs.file
        pages = load_transform(file)

        if pages is None:
            print(f"Unable to load transform from {file}")
            sys.exit(1)

        if not isinstance(pages, list):
            pages = [pages]

        for content in pages:
            args = {
                "path": content["path"],
                "type": content["type"],
                "name": content["name"],
                "implementation": content["implementation"],
                "configuration_schema": content["configurationSchema"],
                "project_uuid": project_uuid,
            }
            for key in ["description", "inputs", "outputs", "tags"]:
                if key in content:
                    args[key] = content[key]

            result = ape_client.update_transform(**args)
            self._print(result)

    @ex(
        help="update an existing generator from a file",
        arguments=[
            (["file"], {"help": "generator definition file", "action": "store"}),
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def generator(self):
        ape_client = client("ape", self.app)

        file = self.app.pargs.file
        if file.endswith(".yaml") or file.endswith(".yml"):
            loader = _yaml_load
        elif file.endswith(".json"):
            loader = json.load
        else:
            print("Only YAML and JSON files may be uploaded.")
            sys.exit(1)

        with open(file, "r", encoding="utf-8") as fd:
            pages = loader(fd)

        if not isinstance(pages, list):
            pages = [pages]

        for content in pages:
            args = {
                "path": content["path"],
                "transform": content["transform"]["path"],
                "transform_configuration": content["transform"]["configuration"],
                "load": content["load"],
                "project_uuid": self._project_scope,
            }

            for key in ["extract", "description", "tags"]:
                if key in content and content[key] is not None:
                    args[key] = content[key]

            result = ape_client.update_generator(**args)
            self._print(result)

    @ex(
        help="update an existing package from a file",
        arguments=[
            (["file"], {"help": "package definition file", "action": "store"}),
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def package(self):
        project_uuid = self._project_scope
        planning_client = client("planning", self.app)

        file = self.app.pargs.file
        with open(file) as fd:
            # read the content of the file
            content = fd.read()

        result = planning_client.update_package(
            project_uuid=project_uuid, content=content
        )
        self._print(result)

    @ex(
        help="update an existing pipeline from a file",
        arguments=[
            (["file"], {"help": "pipeline definition file", "action": "store"}),
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def pipeline(self):
        ape_client = client("ape", self.app)

        file = self.app.pargs.file
        if file.endswith(".yaml") or file.endswith(".yml"):
            loader = _yaml_load
        elif file.endswith(".json"):
            loader = json.load
        else:
            print("Only YAML and JSON files may be uploaded.")
            sys.exit(1)

        with open(file, "r", encoding="utf-8") as fd:
            pages = loader(fd)

        if not isinstance(pages, list):
            pages = [pages]

        for content in pages:
            args = {
                "path": content["path"],
            }
            args["project_uuid"] = self._project_scope

            for key in ["seeders", "generators", "description", "tags"]:
                if key in content and content[key] is not None:
                    args[key] = content[key]

            result = ape_client.update_pipeline(**args)
            self._print(result)

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")



def action(event_name):
    def annotate_action(func):
        func.event_name = event_name
        return func

    return annotate_action


def perception(perception_name):
    def annotate_perception(func):
        func.perception_name = perception_name
        return func

    return annotate_perception


class HandlerMetaclass(type):
    def __new__(cls, name, bases, dct):
        h = super().__new__(cls, name, bases, dct)
        actions = {}
        perceptions = {}
        for name, method in dct.items():
            if hasattr(method, "event_name"):
                actions[method.event_name] = method
            if hasattr(method, "perception_name"):
                perceptions[method.perception_name] = method
        h.actions = actions
        h.perceptions = perceptions
        return h


class Handler(metaclass=HandlerMetaclass):
    pass



import json
import sys

import yaml
from cement import Controller, ex

from ..core.services import client


class Send(Controller):
    """
    Provides the 'send' verb.
    """

    class Meta:
        label = "send"
        stacked_on = "base"
        stacked_type = "nested"
        help = "send something"

    @ex(
        help="send a message to an agent",
        arguments=[
            (["--test", "-t"], {"help": "test mode", "action": "store_true"}),
            (["id"], {"help": "agent instance identifier", "action": "store"}),
            (
                ["events"],
                {"help": "file containing message(s) to send", "action": "store"},
            ),
        ],
    )
    def agent_event(self):
        planning_client = client("planning", self.app)
        file = self.app.pargs.events
        if file == "-":
            # we assume that this is a one-shot set of events, so we read until STDIN closes
            content = sys.stdin.read()
            # read with YAML since that's a superset of JSON
            events = self._yaml_load(content)
        else:
            if file.endswith(".yaml") or file.endswith(".yml"):
                loader = self._yaml_load
            elif file.endswith(".json"):
                loader = json.load
            else:
                print("Only YAML and JSON files may be uploaded.")
                sys.exit(1)
            with open(file, "r", encoding="utf-8") as fd:
                events = loader(fd)

        if not isinstance(events, list):
            events = [events]
        method = planning_client.send_event_to_agent
        if self.app.pargs.test:
            method = planning_client.test_event_with_agent
        for event in events:
            result = method(self.app.pargs.id, event)
            self._print(result)

    @ex(
        help="send a message to an agent swarm",
        arguments=[
            (["id"], {"help": "agent swarm identifier", "action": "store"}),
            (
                ["events"],
                {"help": "file containing message(s) to send", "action": "store"},
            ),
        ],
    )
    def agent_swarm_event(self):
        planning_client = client("planning", self.app)
        file = self.app.pargs.events
        if file == "-":
            # we assume that this is a one-shot set of events, so we read until STDIN closes
            content = sys.stdin.read()
            # read with YAML since that's a superset of JSON
            events = self._yaml_load(content)
        else:
            if file.endswith(".yaml") or file.endswith(".yml"):
                loader = self._yaml_load
            elif file.endswith(".json"):
                loader = json.load
            else:
                print("Only YAML and JSON files may be uploaded.")
                sys.exit(1)
            with open(file, "r", encoding="utf-8") as fd:
                events = loader(fd)

        if not isinstance(events, list):
            events = [events]
        for event in events:
            result = planning_client.send_event_to_agent_swarm(self.app.pargs.id, event)
            self._print(result)

    @staticmethod
    def _yaml_load(fd):
        return list(yaml.load_all(fd, Loader=yaml.Loader))

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")



import json

import yaml
from cement import Controller, ex

from ..core.services import client


class Delete(Controller):
    """
    Provides the 'delete' verb.
    """

    class Meta:
        label = "delete"
        stacked_on = "base"
        stacked_type = "nested"
        help = "delete something"

    @ex(
        help="delete an agent",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["agent"], {"help": "agent uuid or path", "action": "store"}),
        ],
    )
    def agent(self):
        planning_client = client("planning", self.app)
        result = planning_client.delete_agent(
            self.app.pargs.agent, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete an agent swarm",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (
                ["agent_swarm"],
                {"help": "agent swarm uuid or name", "action": "store", "nargs": "?"},
            ),
        ],
    )
    def agent_swarm(self):
        planning_client = client("planning", self.app)
        result = planning_client.delete_agent_swarm(
            self.app.pargs.agent_swarm, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete a transform",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["transform"], {"help": "transform uuid or path", "action": "store"}),
        ],
    )
    def transform(self):
        ape_client = client("ape", self.app)

        result = ape_client.delete_transform(
            self.app.pargs.transform, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete a dataset",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["dataset_uuid"], {"help": "The UUID of the dataset to delete"}),
        ],
    )
    def dataset(self):
        dss_client = client("dss", self.app)
        result = dss_client.delete_dataset(
            self.app.pargs.dataset_uuid, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete a generator",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["generator"], {"help": "generator uuid or path", "action": "store"}),
        ],
    )
    def generator(self):
        ape_client = client("ape", self.app)

        result = ape_client.delete_generator(
            self.app.pargs.generator, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete a package",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["package"], {"help": "package uuid or path", "action": "store"}),
        ],
    )
    def package(self):
        planning_client = client("planning", self.app)
        result = planning_client.delete_package(
            self.app.pargs.package, project_uuid=self._project_scope
        )
        self._print(result)

    @ex(
        help="delete pipeline",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["pipeline"], {"help": "pipeline uuid or path", "action": "store"}),
        ],
    )
    def pipeline(self):
        ape_client = client("ape", self.app)
        pipeline = self.app.pargs.pipeline
        result = ape_client.delete_pipeline(pipeline, project_uuid=self._project_scope)
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")



import json

import yaml
from cement import Controller, ex

from ..core.services import client


class Stop(Controller):
    """
    Provides the 'stop' verb.
    """

    class Meta:
        label = "stop"
        stacked_on = "base"
        stacked_type = "nested"
        help = "stop something"

    @ex(
        help="stop an agent",
        arguments=[(["id"], {"help": "agent instance identifier", "action": "store"})],
    )
    def agent(self):
        planning_client = client("planning", self.app)
        result = planning_client.stop_agent(self.app.pargs.id)
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json.dumps(data))
        else:
            print(yaml.dump(data))



import time

from ursactl.core._base import Base
from ursactl.core.services import client


class PipelineRun(Base):
    """
    Provides access to a pipeline run.
    """

    def __init__(self, *args, pipeline=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = pipeline

    @property
    def client(self):
        if self._client is None:
            self._client = client("ape", self.app)
        return self._client

    @property
    def job_status(self):
        return self._data["jobStatus"]

    @property
    def provenance(self):
        return self._data["provenance"]

    @property
    def parameters(self):
        return self._data["parameters"]

    @property
    def dataset(self):
        if self._data["dataset"]:
            return self.project.Dataset(uuid=self._data["dataset"]["id"])

    def refresh(self):
        if self.uuid is not None:
            self._cached_data = self.client.get_pipeline_run(uuid=self.uuid)

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    "dataset": None,
                    "jobStatus": None,
                    "provenance": None,
                    "parameters": None,
                }
            else:
                self._cached_data = self.client.get_pipeline_run(uuid=self.uuid)
        return self._cached_data

    def __enter__(self):
        while not self.completed:
            time.sleep(5)
            try:
                self.refresh()
            except ConnectionError:
                pass
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @property
    def completed(self):
        return self.job_status == "completed"



from cement import Controller, ex

from ..core.services import client


class Refresh(Controller):
    """
    Provides the 'refresh' verb.
    """

    class Meta:
        label = "refresh"
        stacked_on = "base"
        stacked_type = "nested"
        help = "refresh something"

    @ex(help="refresh token")
    def token(self):
        iam_client = client("iam", self.app)
        iam_client.clear_token()
        iam_client.get_token()



from ursactl.core._base import Base
from ursactl.core.services import client

from .pipeline_run import PipelineRun
from .pipeline_sweep import PipelineSweep


class Pipeline(Base):
    """
    Provides access to a pipeline definition.
    """

    @property
    def client(self):
        if self._client is None:
            self._client = client("ape", self.app)
        return self._client

    @property
    def name(self):
        return self._data["name"]

    @property
    def description(self):
        return self._data["description"]

    @property
    def pipeline_runs(self):
        return []

    @property
    def pipeline_sweeps(self):
        return []

    def run(self, parameters=None):
        """
        Runs the pipeline, returning a PipelineRun object.
        """
        result = self.client.run_pipeline(
            self.uuid, parameters=parameters, project_uuid=self.project.uuid
        )

        return PipelineRun(
            pipeline=self,
            project=self.project,
            uuid=result["id"],
            app=self.app,
            client=self.client,
        )

    def sweep(self, parameters=None, sweep_parameters=None):
        """
        Sweeps the pipeline, returning a PipelineSweep object.
        """
        result = self.client.run_pipeline(
            self.uuid,
            parameters=parameters,
            sweep_parameters=sweep_parameters,
            project_uuid=self.project.uuid,
        )

        return PipelineSweep(
            pipeline=self,
            project=self.project,
            uuid=result["result"]["id"],
            app=self.app,
            client=self.client,
        )

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    "path": None,
                    "description": None,
                    "seeders": [],
                    "generators": [],
                }
            else:
                self._cached_data = self.client.get_pipeline(uuid=self.uuid)
        return self._cached_data



import json
import sys
from pathlib import Path

import yaml
from cement import Controller

import ursactl.controllers.utils.packages
import ursactl.controllers.utils.pages
import ursactl.controllers.utils.transforms

from ..core.services import client
from ..core.utils.magic import magic


class Sync(Controller):
    """
    Provides the 'sync' verb
    """

    class Meta:
        label = "sync"
        stacked_on = "base"
        stacked_type = "nested"
        help = "synchronize a directory with Ursa Frontier"
        arguments = [
            (
                ["--project"],
                {
                    "help": "override default project to use with this repository",
                    "dest": "project",
                    "action": "store",
                },
            ),
            (
                ["--dir", "--directory"],
                {
                    "help": "directory to initialize (defaults to '.')",
                    "dest": "directory",
                    "action": "store",
                    "default": ".",
                },
            ),
            (
                ["--dry-run"],
                {
                    "help": "rather than sync, list all resources that would be synced",
                    "dest": "dry_run",
                    "action": "store_true",
                },
            ),
        ]

    def _default(self):
        root = Path(self.app.pargs.directory)
        if not root.exists():
            print("Error: %s does not exist." % root)
            sys.exit(1)
        elif not root.is_dir():
            print("Error: %s exists and is not a directory." % root)
            sys.exit(1)

        dss_client = client("dss", self.app)
        ape_client = client("ape", self.app)
        page_client = client("page", self.app)
        planning_client = client("planning", self.app)

        resources = self._load_resources(root)
        print(resources)

        if self.app.pargs.dry_run:
            print("Would sync:")
            print(
                yaml.dump(
                    {k: list(v.keys()) for k, v in resources.items() if len(v) > 0}
                )
            )
            return

        # now we can sync resources
        self._sync_pages(page_client, resources["page"])
        self._sync_datasets(dss_client, resources["dataset"])
        self._sync_transforms(ape_client, resources["transform"])
        self._sync_generators(ape_client, resources["generator"])
        self._sync_pipelines(ape_client, resources["pipeline"])
        self._sync_agents(planning_client, resources["agent"])
        self._sync_packages(planning_client, resources["package"])

    def _is_synced(self, collection, path):
        if path not in collection:
            return False
        if "published" in collection[path] and not collection[path]["published"]:
            return False
        return True

    def _sync_datasets(self, client, datasets):
        for item in client.list_datasets(self._project_scope):
            if item["isSynced"]:
                if not self._is_synced(datasets, item["path"]):
                    client.delete_synced_dataset(item["id"])
            elif item["path"] in datasets and not item["isSynced"]:
                print(
                    f"Unable to sync dataset #{item['path']} because it is not synced."
                )
                datasets.delete(item["path"])
        for item in datasets.values():
            args = {
                "path": item["path"],
                "project_uuid": self._project_scope,
                "data": item["file"],
            }
            if "contentType" in item:
                args["data_type"] = item["contentType"]
            else:
                args["data_type"] = magic(item["file"])
            if "description" in item:
                args["description"] = item["description"]
            result = client.sync_dataset(**args)
            self._print(result)

    def _sync_agents(self, client, agents):
        for item in client.list_agents(self._project_scope):
            if item["isSynced"]:
                if not self._is_synced(agents, item["name"]):
                    client.delete_synced_agent(item["id"])
            elif item["name"] in agents and not item["isSynced"]:
                print(f"Unable to sync agent #{item['name']} because it is not synced.")
                agents.delete(item["name"])
        for item in agents.values():
            args = {
                "name": item["name"],
                "project_uuid": self._project_scope,
            }

            for key in ["packages", "behaviors"]:
                if key in item:
                    args[key] = item[key]

            result = client.sync_agent(**args)
            self._print(result)

    def _sync_packages(self, client, packages):
        for item in client.list_packages(self._project_scope):
            if item["isSynced"]:
                if not self._is_synced(packages, item["name"]):
                    client.delete_synced_package(item["id"])
            elif item["name"] in packages and not item["isSynced"]:
                print(
                    f"Unable to sync package #{item['name']} because it is not synced."
                )
                packages.delete(item["name"])
        for item in packages.values():
            args = {
                "project_uuid": self._project_scope,
            }

            args["content"] = item["content"]

            result = client.sync_package(**args)
            self._print(result)

    def _sync_pages(self, client, pages):
        for item in client.list_pages(self._project_scope):
            if item["isSynced"]:
                if not self._is_synced(pages, item["path"]):
                    client.delete_synced_page(item["id"])
            elif item["path"] in pages and not item["isSynced"]:
                print(f"Unable to sync page #{item['path']} because it is not synced.")
                pages.delete(item["path"])
        for item in pages.values():
            args = {
                "path": item["path"],
                "content": item["content"],
                "project_uuid": self._project_scope,
            }

            for key in ["title", "tags"]:
                if key in item:
                    args[key] = item[key]

            result = client.sync_page(**args)
            self._print(result)

    def _sync_transforms(self, client, transforms):
        for item in client.list_transforms(self._project_scope):
            if not self._is_synced(transforms, item["path"]):
                client.delete_synced_transform(item["id"])
            elif item["path"] in transforms and not item["isSynced"]:
                print(
                    f"Unable to sync transform #{item['path']} because it is not synced."
                )
                transforms.delete(item["path"])
        for item in transforms.values():
            args = {
                "path": item["path"],
                "name": item["name"],
                "type": item["type"],
                "implementation": item["implementation"],
                "configuration_schema": item["configurationSchema"],
                "project_uuid": self._project_scope,
            }
            for key in ["outputs", "inputs", "description", "tags"]:
                if key in item:
                    args[key] = item[key]
            result = client.sync_transform(**args)
            self._print(result)

    def _sync_generators(self, client, generators):
        for item in client.list_generators(self._project_scope):
            if not self._is_synced(generators, item["path"]):
                client.delete_synced_generator(item["id"])
            elif item["path"] in generators and not item["isSynced"]:
                print(
                    f"Unable to sync generator #{item['path']} because it is not synced."
                )
                generators.delete(item["path"])
        for item in generators.values():
            args = {
                "path": item["path"],
                "transform": item["transform"]["path"],
                "transform_configuration": item["transform"]["configuration"],
                "load": item["load"],
                "project_uuid": self._project_scope,
            }

            for key in ["extract", "description", "tags"]:
                if key in item and item[key] is not None:
                    args[key] = item[key]

            result = client.sync_generator(**args)
            self._print(result)

    def _sync_pipelines(self, client, pipelines):
        for item in client.list_pipelines(self._project_scope):
            if not self._is_synced(pipelines, item["path"]):
                client.delete_synced_pipeline(item["id"])
            elif item["path"] in pipelines and not item["isSynced"]:
                print(
                    f"Unable to sync pipeline #{item['path']} because it is not synced."
                )
                pipelines.delete(item["path"])
        for item in pipelines.values():
            args = {"path": item["path"], "project_uuid": self._project_scope}
            for key in ["seeders", "generators", "description", "tags"]:
                if key in item and item[key] is not None:
                    args[key] = item[key]

            result = client.sync_pipeline(**args)
            self._print(result)

    def _load_resources(self, root):  # noqa: C901
        # we want to look at each resource we can find locally and compute a hash
        # that makes sense for the data stored on the server - to see if there's a difference
        # for now, we're focused on the data side of things
        # we'll need to look at the planning side of things later
        # we look in each of the directories created by `init`, but use them only to
        # establish the default resource type. Resources can have a `kind` metadata
        # attribute that we can use to determin the kind of resource we're looking at.
        # datasets are special, but the same file with the .yaml extension can house
        # the metadata (so foo.csv has a foo.csv.yaml metadata file)
        #
        # OR, for datasets, we can have an index.yaml file that lists all of the datasets
        # we want to sync along with their metadata. Or both, until we find out which one
        # users find easiest to use.
        resources = {
            "agent": {},
            "package": {},
            "page": {},
            "transform": {},
            "generator": {},
            "pipeline": {},
            "dataset": {},
        }

        for file in root.rglob("*"):
            if file.is_dir():
                continue
            if file.stat().st_size == 0:
                continue
            if file.name.startswith("."):
                continue
            info = None
            suffix = file.suffix
            default_kind = self._get_kind(root, file, {})
            if suffix == ".md":
                # likely a package, transform, or page
                with open(file, "r") as fd:
                    if default_kind == "transform":
                        info = ursactl.controllers.utils.transforms.load_from_markdown(
                            fd
                        )
                    elif default_kind == "package":
                        info = ursactl.controllers.utils.packages.load_from_markdown(fd)
                    else:
                        info = ursactl.controllers.utils.pages.load_from_markdown(fd)
            elif suffix == ".lua":
                # likely a transform
                if file.name.endswith("_spec.lua"):
                    continue  # skip tests
                with open(file, "r") as fd:
                    info = ursactl.controllers.utils.transforms.load_from_lua(fd)
            elif suffix in [".yaml", ".yml"]:
                with open(file, "r") as fd:
                    info = list(yaml.load_all(fd, Loader=yaml.Loader))
            elif suffix == ".json":
                with open(file, "r") as fd:
                    info = json.load(fd)
            elif suffix == ".cog":
                info = ursactl.controllers.utils.packages.load_package(file)
            else:
                continue
            info = info if isinstance(info, list) else [info]
            for item in info:
                kind = self._get_kind(root, file, item)
                if kind in resources:
                    if kind in ["agent", "package"]:
                        resources[kind][item["name"]] = item
                    else:
                        resources[kind][item["path"]] = item
                        if kind == "dataset":
                            resources[kind][item["path"]]["file"] = file.with_name(
                                file.stem
                            )

        return resources

    def _get_kind(self, root, file, info) -> str:
        if info is None:
            return "unknown"
        if "kind" in info and info["kind"] not in [None, ""]:
            return info["kind"]
        if file.suffix in [".yaml", ".yml"] and file.with_name(file.stem).exists():
            return "dataset"
        if file.suffix == ".cog":
            return "package"
        bits = file.relative_to(root).parts
        for type in [
            "agents",
            "packages",
            "pages",
            "transforms",
            "generators",
            "pipelines",
            "datasets",
        ]:
            if type in bits:
                return type[:-1]
        return "unknown"

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get("ursactl", "project")

    def _print(self, data):
        print(yaml.dump(data))



from configparser import NoOptionError
from json import dumps as json_dump

from cement import Controller, ex
from yaml import dump as yaml_dump

from ursactl.core.exc import UrsaProjectNotDefined
from ursactl.core.services import client


class List(Controller):
    """
    Provides the `ursactl list ...` actions.
    """

    class Meta:
        label = "list"
        stacked_on = "base"
        stacked_type = "nested"
        help = "list something"

    @ex(
        help="list agents",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def agents(self):
        planning_client = client("planning", self.app)

        agents = planning_client.list_agents(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in agents:
                data.append([item["name"]])
            data.sort()
            self.app.render(data, headers=["Name"])
        else:
            self._print(agents)

    @ex(
        help="list agent swarms",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def agent_swarms(self):
        planning_client = client("planning", self.app)

        swarms = planning_client.list_agent_swarms(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in swarms:
                data.append([item["name"], item["id"]])
            data.sort()
            self.app.render(data, headers=["Name", "ID"])
        else:
            self._print(swarms)

    @ex(
        help="list transforms",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def transforms(self):
        ape_client = client("ape", self.app)

        algos = ape_client.list_transforms(project_uuid=self._project_scope)

        if self.app.pargs.output is None:
            data = []
            for item in algos:
                data.append([item["path"], item["name"], item["type"]])
            data.sort()
            self.app.render(data, headers=["Path", "Name", "Type"])
        else:
            self._print(algos)

    @ex(
        help="list datasets",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def datasets(self):
        dss_client = client("dss", self.app)
        datasets = dss_client.list_datasets(self._project_scope)

        if self.app.pargs.output is None:
            data = []
            for item in datasets:
                data.append([item["path"], item["size"], item["contentType"]])
            data.sort()
            self.app.render(data, headers=["Path", "Size", "Type"])
        else:
            self._print(list(datasets))

    @ex(
        help="list pages",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def pages(self):
        page_client = client("page", self.app)

        pages = page_client.list_pages(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in pages:
                data.append([item["path"], item["title"]])
            data.sort()
            self.app.render(data, headers=["Path", "Title"])
        else:
            self._print(pages)

    @ex(
        help="list generators",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def generators(self):
        ape_client = client("ape", self.app)

        generators = ape_client.list_generators(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in generators:
                data.append(
                    [
                        item["path"],
                        item["transform"]["type"],
                        self._transform_name(item["transform"]),
                    ]
                )
            data.sort()
            self.app.render(data, headers=["Path", "Type", "transform"])
        else:
            self._print(generators)

    @ex(
        help="list packages",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def packages(self):
        planning_client = client("planning", self.app)

        packages = planning_client.list_packages(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in packages:
                data.append([item["name"]])
            data.sort()
            self.app.render(data, headers=["Name"])
        else:
            self._print(packages)

    @ex(
        help="list pipelines",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
        ],
    )
    def pipelines(self):
        ape_client = client("ape", self.app)

        if self.app.pargs.output is None:
            pipelines = ape_client.list_pipelines(project_uuid=self._project_scope)
            data = []
            for item in pipelines:
                data.append([item["path"], item["description"]])
            data.sort()
            self.app.render(data, headers=["Path", "Description"])
        else:
            pipelines = ape_client.list_pipelines(
                project_uuid=self._project_scope, all=True
            )
            self._print(pipelines)

    @ex(help="list projects")
    def projects(self):
        projects_client = client("projects", self.app)
        projects = projects_client.list()

        if self.app.pargs.output is None:
            data = []
            for item in projects:
                data.append(
                    [
                        item["name"],
                        item["isArchived"],
                        item["user"]["handle"],
                        item["description"],
                    ]
                )
            data.sort()
            self.app.render(data, headers=["Name", "Archived?", "Owner", "Description"])
        else:
            self._print(list(projects))

    @ex(
        help="list running agents",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (["agent"], {"help": "agent identifier", "action": "store", "nargs": "?"}),
        ],
    )
    def running_agents(self):
        planning_client = client("planning", self.app)

        running_agents = planning_client.list_running_agents(
            agent_id=self.app.pargs.agent, project_uuid=self._project_scope
        )
        if self.app.pargs.output is None:
            data = []
            for item in running_agents:
                data.append(
                    [item["agent"], item["startedAt"], item["id"], item["status"]]
                )
            data.sort()
            self.app.render(
                data, headers=["Agent", "Started At", "Instance UUID", "Status"]
            )
        else:
            self._print(running_agents)

    @ex(
        help="list pipeline runs",
        arguments=[
            (
                ["--project"],
                {"help": "project identifier", "dest": "project", "action": "store"},
            ),
            (
                ["pipeline"],
                {"help": "pipeline identifier", "action": "store", "nargs": "?"},
            ),
        ],
    )
    def pipeline_runs(self):
        ape_client = client("ape", self.app)

        pipeline_runs = ape_client.list_pipeline_runs(
            pipeline_id=self.app.pargs.pipeline, project_uuid=self._project_scope
        )

        if self.app.pargs.output is None:
            data = []
            for item in pipeline_runs:
                data.append([item["pipeline"], item["jobStatus"], item["dataset"]])
            data.sort()
            self.app.render(data, headers=["Pipeline", "Status", "Dataset"])
        else:
            self._print(pipeline_runs)

    @property
    def _project_scope(self):
        try:
            scope = self.app.pargs.project or self.app.config.get("ursactl", "project")
        except NoOptionError:
            raise UrsaProjectNotDefined("Project must be defined.")
        return scope

    def _print(self, data):
        if self.app.pargs.output == "json":
            print(json_dump(data))
        else:
            print(yaml_dump(data))

    def _transform_name(self, transform):
        return f"{transform['project']['user']['handle']}/{transform['project']['name']}:{transform['path']}"



import asyncio
import json
from urllib.parse import urlencode

import websockets

from .messages import IncomingMessage
from .utils import DONE, get_unless_done


def _load_incoming_message(data):
    return IncomingMessage(*data)


def dump_outgoing_message(message):
    return [
        message.join_ref,
        message.ref,
        message.topic,
        message.event,
        message.payload,
    ]


class Transport:
    def __init__(self, url, params, incoming_queue, outgoing_queue):
        self.incoming_queue = incoming_queue
        self.outgoing_queue = outgoing_queue
        qs_params = {"vsn": "2.0.0", **params}
        self.url = url + "?" + urlencode(qs_params)
        self.ready = asyncio.Future()
        self._done = asyncio.Future()

    async def run(self):
        try:
            async with websockets.connect(self.url) as websocket:
                self.ready.set_result(True)
                await asyncio.gather(
                    self._recv_loop(websocket), self._send_loop(websocket)
                )
        except Exception as e:
            if not self.ready.done():
                self.ready.set_exception(e)

            raise

    async def stop(self):
        self._done.set_result(True)
        return True

    async def _recv_loop(self, websocket):
        while True:
            message_data = await get_unless_done(websocket.recv(), self._done)
            if message_data is DONE:
                return

            message = _load_incoming_message(json.loads(message_data))
            await self.incoming_queue.put(message)

    async def _send_loop(self, websocket):
        while True:
            message = await get_unless_done(self.outgoing_queue.get(), self._done)

            if message is DONE:
                return

            try:
                message_data = json.dumps(dump_outgoing_message(message))

                await websocket.send(message_data)
                message.sent.set_result(True)
            except Exception as e:
                print(e)
                message.sent.set_exception(e)
